@@ -22,9 +22,7 @@
 #ifndef AWESOME_BANNING_H
 #define AWESOME_BANNING_H
 
-#include "globalconf.h"
-
-void banning_need_update(screen_t *);
+void banning_need_update(void);
 void banning_refresh(void);
 
 #endif
