@@ -22,7 +22,6 @@
 #ifndef AWESOME_COMMON_SIGNAL
 #define AWESOME_COMMON_SIGNAL
 
-#include "common/lualib.h"
 #include "common/array.h"
 
 DO_ARRAY(const void *, cptr, DO_NOTHING)
@@ -55,23 +54,6 @@ signal_array_getbyid(signal_array_t *arr, unsigned long id)
     return signal_array_lookup(arr, &sig);
 }
 
-/** Add a signal to a signal array.
- * Signals have to be added before they can be added or emitted.
- * \param arr The signal array.
- * \param name The signal name.
- */
-static inline void
-signal_add(signal_array_t *arr, const char *name)
-{
-    unsigned long tok = a_strhash((const unsigned char *) name);
-    signal_t *sigfound = signal_array_getbyid(arr, tok);
-    if(!sigfound)
-    {
-        signal_t sig = { .id = tok };
-        signal_array_insert(arr, sig);
-    }
-}
-
 /** Connect a signal inside a signal array.
  * You are in charge of reference counting.
  * \param arr The signal array.
@@ -86,7 +68,11 @@ signal_connect(signal_array_t *arr, const char *name, const void *ref)
     if(sigfound)
         cptr_array_append(&sigfound->sigfuncs, ref);
     else
-        warn("Trying to connect to unknown signal '%s'", name);
+    {
+        signal_t sig = { .id = tok };
+        cptr_array_append(&sig.sigfuncs, ref);
+        signal_array_insert(arr, sig);
+    }
 }
 
 /** Disconnect a signal inside a signal array.
@@ -95,7 +81,7 @@ signal_connect(signal_array_t *arr, const char *name, const void *ref)
  * \param name The signal name.
  * \param ref The reference to remove.
  */
-static inline void
+static inline bool
 signal_disconnect(signal_array_t *arr, const char *name, const void *ref)
 {
     signal_t *sigfound = signal_array_getbyid(arr,
@@ -106,10 +92,12 @@ signal_disconnect(signal_array_t *arr, const char *name, const void *ref)
             if(ref == *func)
             {
                 cptr_array_remove(&sigfound->sigfuncs, func);
-                break;
+                if(sigfound->sigfuncs.len == 0)
+                    signal_array_remove(arr, sigfound);
+                return true;
             }
-    } else
-        warn("Trying to disconnect from unknown signal '%s'", name);
+    }
+    return false;
 }
 
 #endif
