@@ -19,9 +19,15 @@
  *
  */
 
-#include "config.h"
+/** awesome D-Bus API
+ * @author Julien Danjou &lt;julien@danjou.info&gt;
+ * @copyright 2008-2009 Julien Danjou
+ * @module dbus
+ */
 
+#include "config.h"
 #include "dbus.h"
+
 #include <glib.h>
 
 #ifdef WITH_DBUS
@@ -42,7 +48,7 @@ static signal_array_t dbus_signals;
 
 /** Clean up the D-Bus connection data members
  * \param dbus_connection The D-Bus connection to clean up
- * \param dbusio The D-Bus event watcher
+ * \param source The D-Bus source
  */
 static void
 a_dbus_cleanup_bus(DBusConnection *dbus_connection, GSource **source)
@@ -61,11 +67,12 @@ a_dbus_cleanup_bus(DBusConnection *dbus_connection, GSource **source)
 }
 
 /** Iterate through the D-Bus messages counting each or traverse each sub message.
+ * \param L The Lua VM state.
  * \param iter The D-Bus message iterator pointer
  * \return The number of arguments in the iterator
  */
 static int
-a_dbus_message_iter(DBusMessageIter *iter)
+a_dbus_message_iter(lua_State *L, DBusMessageIter *iter)
 {
     int nargs = 0;
 
@@ -74,7 +81,7 @@ a_dbus_message_iter(DBusMessageIter *iter)
         switch(dbus_message_iter_get_arg_type(iter))
         {
           default:
-            lua_pushnil(globalconf.L);
+            lua_pushnil(L);
             nargs++;
             break;
           case DBUS_TYPE_INVALID:
@@ -83,7 +90,7 @@ a_dbus_message_iter(DBusMessageIter *iter)
             {
                 DBusMessageIter subiter;
                 dbus_message_iter_recurse(iter, &subiter);
-                a_dbus_message_iter(&subiter);
+                a_dbus_message_iter(L, &subiter);
             }
             nargs++;
             break;
@@ -94,7 +101,7 @@ a_dbus_message_iter(DBusMessageIter *iter)
                 /* initialize a sub iterator */
                 dbus_message_iter_recurse(iter, &subiter);
                 /* create a new table to store the dict */
-                a_dbus_message_iter(&subiter);
+                a_dbus_message_iter(L, &subiter);
             }
             nargs++;
             break;
@@ -104,15 +111,15 @@ a_dbus_message_iter(DBusMessageIter *iter)
                 /* initialize a sub iterator */
                 dbus_message_iter_recurse(iter, &subiter);
 
-                int n = a_dbus_message_iter(&subiter);
+                int n = a_dbus_message_iter(L, &subiter);
 
                 /* create a new table to store all the value */
-                lua_createtable(globalconf.L, n, 0);
+                lua_createtable(L, n, 0);
                 /* move the table before array elements */
-                lua_insert(globalconf.L, - n - 1);
+                lua_insert(L, - n - 1);
 
                 for(int i = n; i > 0; i--)
-                    lua_rawseti(globalconf.L, - i - 1, i);
+                    lua_rawseti(L, - i - 1, i);
             }
             nargs++;
             break;
@@ -127,44 +134,44 @@ a_dbus_message_iter(DBusMessageIter *iter)
 
                     switch(array_type)
                     {
-                      int datalen = 0;
-#define DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER(type, dbustype) \
+                      int datalen;
+#define DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER_OR_INT(type, dbustype, pusher) \
                       case dbustype: \
                         { \
                             const type *data; \
                             dbus_message_iter_get_fixed_array(&sub, &data, &datalen); \
-                            lua_createtable(globalconf.L, datalen, 0); \
+                            lua_createtable(L, datalen, 0); \
                             for(int i = 0; i < datalen; i++) \
                             { \
-                                lua_pushnumber(globalconf.L, data[i]); \
-                                lua_rawseti(globalconf.L, -2, i + 1); \
+                                pusher(L, data[i]); \
+                                lua_rawseti(L, -2, i + 1); \
                             } \
                         } \
                         break;
-                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER(int16_t, DBUS_TYPE_INT16)
-                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER(uint16_t, DBUS_TYPE_UINT16)
-                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER(int32_t, DBUS_TYPE_INT32)
-                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER(uint32_t, DBUS_TYPE_UINT32)
-                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER(int64_t, DBUS_TYPE_INT64)
-                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER(uint64_t, DBUS_TYPE_UINT64)
-                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER(double, DBUS_TYPE_DOUBLE)
-#undef DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER
+                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER_OR_INT(int16_t, DBUS_TYPE_INT16, lua_pushinteger)
+                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER_OR_INT(uint16_t, DBUS_TYPE_UINT16, lua_pushinteger)
+                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER_OR_INT(int32_t, DBUS_TYPE_INT32, lua_pushinteger)
+                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER_OR_INT(uint32_t, DBUS_TYPE_UINT32, lua_pushinteger)
+                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER_OR_INT(int64_t, DBUS_TYPE_INT64, lua_pushinteger)
+                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER_OR_INT(uint64_t, DBUS_TYPE_UINT64, lua_pushinteger)
+                      DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER_OR_INT(double, DBUS_TYPE_DOUBLE, lua_pushnumber)
+#undef DBUS_MSG_HANDLE_ARRAY_TYPE_NUMBER_OR_INT
                       case DBUS_TYPE_BYTE:
                         {
                             const char *c;
                             dbus_message_iter_get_fixed_array(&sub, &c, &datalen);
-                            lua_pushlstring(globalconf.L, c, datalen);
+                            lua_pushlstring(L, c, datalen);
                         }
                         break;
                       case DBUS_TYPE_BOOLEAN:
                         {
                             const dbus_bool_t *b;
                             dbus_message_iter_get_fixed_array(&sub, &b, &datalen);
-                            lua_createtable(globalconf.L, datalen, 0);
+                            lua_createtable(L, datalen, 0);
                             for(int i = 0; i < datalen; i++)
                             {
-                                lua_pushboolean(globalconf.L, b[i]);
-                                lua_rawseti(globalconf.L, -2, i + 1);
+                                lua_pushboolean(L, b[i]);
+                                lua_rawseti(L, -2, i + 1);
                             }
                         }
                         break;
@@ -178,15 +185,15 @@ a_dbus_message_iter(DBusMessageIter *iter)
 
                     /* get the keys and the values
                      * n is the number of entry in dict */
-                    int n = a_dbus_message_iter(&subiter);
+                    int n = a_dbus_message_iter(L, &subiter);
 
                     /* create a new table to store all the value */
-                    lua_createtable(globalconf.L, n, 0);
+                    lua_createtable(L, n, 0);
                     /* move the table before array elements */
-                    lua_insert(globalconf.L, - (n * 2) - 1);
+                    lua_insert(L, - (n * 2) - 1);
 
                     for(int i = 0; i < n; i ++)
-                        lua_rawset(globalconf.L, - (n * 2) - 1 + i * 2);
+                        lua_rawset(L, - (n * 2) - 1 + i * 2);
                 }
                 else
                 {
@@ -195,15 +202,15 @@ a_dbus_message_iter(DBusMessageIter *iter)
                     dbus_message_iter_recurse(iter, &subiter);
 
                     /* now iterate over every element of the array */
-                    int n = a_dbus_message_iter(&subiter);
+                    int n = a_dbus_message_iter(L, &subiter);
 
                     /* create a new table to store all the value */
-                    lua_createtable(globalconf.L, n, 0);
+                    lua_createtable(L, n, 0);
                     /* move the table before array elements */
-                    lua_insert(globalconf.L, - n - 1);
+                    lua_insert(L, - n - 1);
 
                     for(int i = n; i > 0; i--)
-                        lua_rawseti(globalconf.L, - i - 1, i);
+                        lua_rawseti(L, - i - 1, i);
                 }
             }
             nargs++;
@@ -212,7 +219,7 @@ a_dbus_message_iter(DBusMessageIter *iter)
             {
                 dbus_bool_t b;
                 dbus_message_iter_get_basic(iter, &b);
-                lua_pushboolean(globalconf.L, b);
+                lua_pushboolean(L, b);
             }
             nargs++;
             break;
@@ -220,31 +227,32 @@ a_dbus_message_iter(DBusMessageIter *iter)
             {
                 char c;
                 dbus_message_iter_get_basic(iter, &c);
-                lua_pushlstring(globalconf.L, &c, 1);
+                lua_pushlstring(L, &c, 1);
             }
             nargs++;
             break;
-#define DBUS_MSG_HANDLE_TYPE_NUMBER(type, dbustype) \
+#define DBUS_MSG_HANDLE_TYPE_NUMBER_OR_INT(type, dbustype, pusher) \
           case dbustype: \
             { \
                 type ui; \
                 dbus_message_iter_get_basic(iter, &ui); \
-                lua_pushnumber(globalconf.L, ui); \
+                pusher(L, ui); \
             } \
             nargs++; \
             break;
-          DBUS_MSG_HANDLE_TYPE_NUMBER(int16_t, DBUS_TYPE_INT16)
-          DBUS_MSG_HANDLE_TYPE_NUMBER(uint16_t, DBUS_TYPE_UINT16)
-          DBUS_MSG_HANDLE_TYPE_NUMBER(int32_t, DBUS_TYPE_INT32)
-          DBUS_MSG_HANDLE_TYPE_NUMBER(uint32_t, DBUS_TYPE_UINT32)
-          DBUS_MSG_HANDLE_TYPE_NUMBER(int64_t, DBUS_TYPE_INT64)
-          DBUS_MSG_HANDLE_TYPE_NUMBER(uint64_t, DBUS_TYPE_UINT64)
-#undef DBUS_MSG_HANDLE_TYPE_NUMBER
+          DBUS_MSG_HANDLE_TYPE_NUMBER_OR_INT(int16_t, DBUS_TYPE_INT16, lua_pushinteger)
+          DBUS_MSG_HANDLE_TYPE_NUMBER_OR_INT(uint16_t, DBUS_TYPE_UINT16, lua_pushinteger)
+          DBUS_MSG_HANDLE_TYPE_NUMBER_OR_INT(int32_t, DBUS_TYPE_INT32, lua_pushinteger)
+          DBUS_MSG_HANDLE_TYPE_NUMBER_OR_INT(uint32_t, DBUS_TYPE_UINT32, lua_pushinteger)
+          DBUS_MSG_HANDLE_TYPE_NUMBER_OR_INT(int64_t, DBUS_TYPE_INT64, lua_pushinteger)
+          DBUS_MSG_HANDLE_TYPE_NUMBER_OR_INT(uint64_t, DBUS_TYPE_UINT64, lua_pushinteger)
+          DBUS_MSG_HANDLE_TYPE_NUMBER_OR_INT(double, DBUS_TYPE_DOUBLE, lua_pushnumber)
+#undef DBUS_MSG_HANDLE_TYPE_NUMBER_OR_INT
           case DBUS_TYPE_STRING:
             {
                 char *s;
                 dbus_message_iter_get_basic(iter, &s);
-                lua_pushstring(globalconf.L, s);
+                lua_pushstring(L, s);
             }
             nargs++;
             break;
@@ -259,7 +267,7 @@ a_dbus_convert_value(lua_State *L, int idx, DBusMessageIter *iter)
 {
     /* i is the type name, i+1 the value */
     size_t len;
-    const char *type = lua_tolstring(globalconf.L, idx, &len);
+    const char *type = lua_tolstring(L, idx, &len);
 
     if(!type || len < 1)
         return false;
@@ -277,7 +285,7 @@ a_dbus_convert_value(lua_State *L, int idx, DBusMessageIter *iter)
 
             if(arraylen % 2 != 0)
             {
-                luaA_warn(globalconf.L,
+                luaA_warn(L,
                           "your D-Bus signal handling method returned wrong number of arguments");
                 return false;
             }
@@ -302,28 +310,28 @@ a_dbus_convert_value(lua_State *L, int idx, DBusMessageIter *iter)
         break;
       case DBUS_TYPE_BOOLEAN:
         {
-            dbus_bool_t b = lua_toboolean(globalconf.L, idx + 1);
+            dbus_bool_t b = lua_toboolean(L, idx + 1);
             dbus_message_iter_append_basic(iter, DBUS_TYPE_BOOLEAN, &b);
         }
         break;
-#define DBUS_MSG_RETURN_HANDLE_TYPE_STRING(dbustype) \
-      case dbustype: \
-        { \
-            const char *s = lua_tostring(globalconf.L, idx + 1); \
-            if(s) \
-                dbus_message_iter_append_basic(iter, dbustype, &s); \
-        } \
+      case DBUS_TYPE_STRING:
+        {
+            const char *s = lua_tostring(L, idx + 1);
+            if(!s || !dbus_validate_utf8(s, NULL)) {
+                luaA_warn(L, "Your D-Bus signal handling method returned an invalid string");
+                return false;
+            }
+            dbus_message_iter_append_basic(iter, DBUS_TYPE_STRING, &s);
+        }
         break;
-    DBUS_MSG_RETURN_HANDLE_TYPE_STRING(DBUS_TYPE_STRING)
-    DBUS_MSG_RETURN_HANDLE_TYPE_STRING(DBUS_TYPE_BYTE)
-#undef DBUS_MSG_RETURN_HANDLE_TYPE_STRING
 #define DBUS_MSG_RETURN_HANDLE_TYPE_NUMBER(type, dbustype) \
       case dbustype: \
         { \
-           type num = lua_tonumber(globalconf.L, idx + 1); \
+           type num = lua_tonumber(L, idx + 1); \
            dbus_message_iter_append_basic(iter, dbustype, &num); \
         } \
         break;
+    DBUS_MSG_RETURN_HANDLE_TYPE_NUMBER(uint8_t, DBUS_TYPE_BYTE)
     DBUS_MSG_RETURN_HANDLE_TYPE_NUMBER(int16_t, DBUS_TYPE_INT16)
     DBUS_MSG_RETURN_HANDLE_TYPE_NUMBER(uint16_t, DBUS_TYPE_UINT16)
     DBUS_MSG_RETURN_HANDLE_TYPE_NUMBER(int32_t, DBUS_TYPE_INT32)
@@ -345,54 +353,61 @@ static void
 a_dbus_process_request(DBusConnection *dbus_connection, DBusMessage *msg)
 {
     const char *interface = dbus_message_get_interface(msg);
-    int old_top = lua_gettop(globalconf.L);
+    lua_State *L = globalconf_get_lua_State();
+    int old_top = lua_gettop(L);
 
-    lua_createtable(globalconf.L, 0, 5);
+    lua_createtable(L, 0, 5);
 
     switch(dbus_message_get_type(msg))
     {
       case DBUS_MESSAGE_TYPE_SIGNAL:
-        lua_pushliteral(globalconf.L, "signal");
+        lua_pushliteral(L, "signal");
         break;
       case DBUS_MESSAGE_TYPE_METHOD_CALL:
-        lua_pushliteral(globalconf.L, "method_call");
+        lua_pushliteral(L, "method_call");
         break;
       case DBUS_MESSAGE_TYPE_METHOD_RETURN:
-        lua_pushliteral(globalconf.L, "method_return");
+        lua_pushliteral(L, "method_return");
         break;
       case DBUS_MESSAGE_TYPE_ERROR:
-        lua_pushliteral(globalconf.L, "error");
+        lua_pushliteral(L, "error");
         break;
       default:
-        lua_pushliteral(globalconf.L, "unknown");
+        lua_pushliteral(L, "unknown");
         break;
     }
 
-    lua_setfield(globalconf.L, -2, "type");
+    lua_setfield(L, -2, "type");
 
-    lua_pushstring(globalconf.L, interface);
-    lua_setfield(globalconf.L, -2, "interface");
+    lua_pushstring(L, interface);
+    lua_setfield(L, -2, "interface");
 
     const char *s = dbus_message_get_path(msg);
-    lua_pushstring(globalconf.L, s);
-    lua_setfield(globalconf.L, -2, "path");
+    lua_pushstring(L, s);
+    lua_setfield(L, -2, "path");
 
     s = dbus_message_get_member(msg);
-    lua_pushstring(globalconf.L, s);
-    lua_setfield(globalconf.L, -2, "member");
+    lua_pushstring(L, s);
+    lua_setfield(L, -2, "member");
+
+    s = dbus_message_get_sender(msg);
+    if(s != NULL) {
+        lua_pushstring(L, s);
+        lua_setfield(L, -2, "sender");
+    }
 
     if(dbus_connection == dbus_connection_system)
-        lua_pushliteral(globalconf.L, "system");
+        lua_pushliteral(L, "system");
     else
-        lua_pushliteral(globalconf.L, "session");
-    lua_setfield(globalconf.L, -2, "bus");
+        lua_pushliteral(L, "session");
+    lua_setfield(L, -2, "bus");
 
     /* + 1 for the table above */
     DBusMessageIter iter;
     int nargs = 1;
 
     if(dbus_message_iter_init(msg, &iter))
-        nargs += a_dbus_message_iter(&iter);
+        nargs += a_dbus_message_iter(L, &iter);
 
     if(dbus_message_get_no_reply(msg))
     {
@@ -400,7 +415,7 @@ a_dbus_process_request(DBusConnection *dbus_connection, DBusMessage *msg)
                                                   a_strhash((const unsigned char *) NONULL(interface)));
         /* emit signals */
         if(sigfound)
-            signal_object_emit(globalconf.L, &dbus_signals, NONULL(interface), nargs);
+            signal_object_emit(L, &dbus_signals, NONULL(interface), nargs);
     }
     else
     {
@@ -411,12 +426,12 @@ a_dbus_process_request(DBusConnection *dbus_connection, DBusMessage *msg)
             /* there can be only ONE handler to send reply */
             void *func = (void *) sig->sigfuncs.tab[0];
 
-            int n = lua_gettop(globalconf.L) - nargs;
+            int n = lua_gettop(L) - nargs;
 
-            luaA_object_push(globalconf.L, (void *) func);
-            luaA_dofunction(globalconf.L, nargs, LUA_MULTRET);
+            luaA_object_push(L, (void *) func);
+            luaA_dofunction(L, nargs, LUA_MULTRET);
 
-            n -= lua_gettop(globalconf.L);
+            n -= lua_gettop(L);
 
             DBusMessage *reply = dbus_message_new_method_return(msg);
 
@@ -424,26 +439,26 @@ a_dbus_process_request(DBusConnection *dbus_connection, DBusMessage *msg)
 
             if(n % 2 != 0)
             {
-                luaA_warn(globalconf.L,
+                luaA_warn(L,
                           "your D-Bus signal handling method returned wrong number of arguments");
                 /* Restore stack */
-                lua_settop(globalconf.L, old_top);
+                lua_settop(L, old_top);
                 return;
             }
 
             /* i is negative */
             for(int i = n; i < 0; i += 2)
             {
-                if(!a_dbus_convert_value(globalconf.L, i, &iter))
+                if(!a_dbus_convert_value(L, i, &iter))
                 {
-                    luaA_warn(globalconf.L, "your D-Bus signal handling method returned bad data");
+                    luaA_warn(L, "your D-Bus signal handling method returned bad data");
                     /* Restore stack */
-                    lua_settop(globalconf.L, old_top);
+                    lua_settop(L, old_top);
                     return;
                 }
 
-                lua_remove(globalconf.L, i);
-                lua_remove(globalconf.L, i + 1);
+                lua_remove(L, i);
+                lua_remove(L, i + 1);
             }
 
             dbus_connection_send(dbus_connection, reply, NULL);
@@ -451,12 +466,12 @@ a_dbus_process_request(DBusConnection *dbus_connection, DBusMessage *msg)
         }
     }
     /* Restore stack */
-    lua_settop(globalconf.L, old_top);
+    lua_settop(L, old_top);
 }
 
 /** Attempt to process all the requests in the D-Bus connection.
  * \param dbus_connection The D-Bus connection to process from
- * \param dbusio The D-Bus event watcher
+ * \param source The D-Bus source
  */
 static void
 a_dbus_process_requests_on_bus(DBusConnection *dbus_connection, GSource **source)
@@ -489,10 +504,6 @@ a_dbus_process_requests_on_bus(DBusConnection *dbus_connection, GSource **source
         dbus_connection_flush(dbus_connection);
 }
 
-/** Foreword D-Bus process session requests on too the correct function.
- * \param w The D-Bus event watcher
- * \param revents (not used)
- */
 static gboolean
 a_dbus_process_requests_session(gpointer data)
 {
@@ -500,10 +511,6 @@ a_dbus_process_requests_session(gpointer data)
     return TRUE;
 }
 
-/** Foreword D-Bus process system requests on too the correct function.
- * \param w The D-Bus event watcher
- * \param revents (not used)
- */
 static gboolean
 a_dbus_process_requests_system(gpointer data)
 {
@@ -511,9 +518,9 @@ a_dbus_process_requests_system(gpointer data)
     return TRUE;
 }
 
-/** Attempt to request a D-Bus name
- * \param dbus_connection The application's connection to D-Bus
- * \param name The D-Bus connection name to be requested
+/** Attempt to request a D-Bus name.
+ * \param dbus_connection The application's connection to D-Bus.
+ * \param name The D-Bus connection name to be requested.
  * \return true if the name is primary owner or the name is already
  * the owner, otherwise false.
  */
@@ -603,7 +610,7 @@ a_dbus_connect(DBusBusType type, const char *type_name, GSourceFunc cb, GSource
     dbus_connection = dbus_bus_get(type, &err);
     if(dbus_error_is_set(&err))
     {
-        warn("D-Bus session bus %s failed: %s", type_name, err.message);
+        warn("Could not connect to D-Bus %s bus: %s", type_name, err.message);
         dbus_connection = NULL;
         dbus_error_free(&err);
     }
@@ -650,9 +657,9 @@ a_dbus_cleanup(void)
     a_dbus_cleanup_bus(dbus_connection_system, &system_source);
 }
 
-/** Retrieve the D-Bus bus by it's name
- * \param name The name of the bus
- * \return The corresponding D-Bus connection
+/** Retrieve the D-Bus bus by its name.
+ * \param name The name of the bus.
+ * \return The corresponding D-Bus connection.
  */
 static DBusConnection *
 a_dbus_bus_getbyname(const char *name)
@@ -664,13 +671,12 @@ a_dbus_bus_getbyname(const char *name)
     return NULL;
 }
 
-/** Register a D-Bus name to receive message from.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- * \luastack
- * \lparam A string indicating if we are using system or session bus.
- * \lparam A string with the name of the D-Bus name to register.
- * \lreturn True if everything worked fine, false otherwise.
+/** Register a D-Bus name to receive messages from.
+ *
+ * @param bus A string indicating if we are using system or session bus.
+ * @param name A string with the name of the D-Bus name to register.
+ * @return True if everything worked fine, false otherwise.
+ * @function request_name
  */
 static int
 luaA_dbus_request_name(lua_State *L)
@@ -683,12 +689,11 @@ luaA_dbus_request_name(lua_State *L)
 }
 
 /** Release a D-Bus name.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- * \luastack
- * \lparam A string indicating if we are using system or session bus.
- * \lparam A string with the name of the D-Bus name to unregister.
- * \lreturn True if everything worked fine, false otherwise.
+ *
+ * @param bus A string indicating if we are using system or session bus.
+ * @param name A string with the name of the D-Bus name to unregister.
+ * @return True if everything worked fine, false otherwise.
+ * @function release_name
  */
 static int
 luaA_dbus_release_name(lua_State *L)
@@ -701,11 +706,10 @@ luaA_dbus_release_name(lua_State *L)
 }
 
 /** Add a match rule to match messages going through the message bus.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- * \luastack
- * \lparam A string indicating if we are using system or session bus.
- * \lparam A string with the name of the match rule.
+ *
+ * @param bus A string indicating if we are using system or session bus.
+ * @param name A string with the name of the match rule.
+ * @function add_match
  */
 static int
 luaA_dbus_add_match(lua_State *L)
@@ -725,11 +729,10 @@ luaA_dbus_add_match(lua_State *L)
 
 /** Remove a previously added match rule "by value"
  * (the most recently-added identical rule gets removed).
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- * \luastack
- * \lparam A string indicating if we are using system or session bus.
- * \lparam A string with the name of the match rule.
+ *
+ * @param bus A string indicating if we are using system or session bus.
+ * @param name A string with the name of the match rule.
+ * @function remove_match
  */
 static int
 luaA_dbus_remove_match(lua_State *L)
@@ -748,11 +751,12 @@ luaA_dbus_remove_match(lua_State *L)
 }
 
 /** Add a signal receiver on the D-Bus.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- * \luastack
- * \lparam A string with the interface name.
- * \lparam The function to call.
+ *
+ * @param interface A string with the interface name.
+ * @param func The function to call.
+ * @return true on success, nil + error if the signal could not be connected
+ * because another function is already connected.
+ * @function connect_signal
  */
 static int
 luaA_dbus_connect_signal(lua_State *L)
@@ -761,22 +765,23 @@ luaA_dbus_connect_signal(lua_State *L)
     luaA_checkfunction(L, 2);
     signal_t *sig = signal_array_getbyid(&dbus_signals,
                                          a_strhash((const unsigned char *) name));
-    if(sig)
+    if(sig) {
         luaA_warn(L, "cannot add signal %s on D-Bus, already existing", name);
-    else
-    {
-        signal_add(&dbus_signals, name);
+        lua_pushnil(L);
+        lua_pushfstring(L, "cannot add signal %s on D-Bus, already existing", name);
+        return 2;
+    } else {
         signal_connect(&dbus_signals, name, luaA_object_ref(L, 2));
+        lua_pushboolean(L, 1);
+        return 1;
     }
-    return 0;
 }
 
-/** Add a signal receiver on the D-Bus.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- * \luastack
- * \lparam A string with the interface name.
- * \lparam The function to call.
+/** Remove a signal receiver on the D-Bus.
+ *
+ * @param interface A string with the interface name.
+ * @param func The function to call.
+ * @function disconnect_signal
  */
 static int
 luaA_dbus_disconnect_signal(lua_State *L)
@@ -784,11 +789,66 @@ luaA_dbus_disconnect_signal(lua_State *L)
     const char *name = luaL_checkstring(L, 1);
     luaA_checkfunction(L, 2);
     const void *func = lua_topointer(L, 2);
-    signal_disconnect(&dbus_signals, name, func);
-    luaA_object_unref(L, func);
+    if (signal_disconnect(&dbus_signals, name, func))
+        luaA_object_unref(L, func);
     return 0;
 }
 
+/** Emit a signal on the D-Bus.
+ *
+ * @param bus A string indicating if we are using system or session bus.
+ * @param path A string with the dbus path.
+ * @param interface A string with the dbus interface.
+ * @param method A string with the dbus method name.
+ * @param type_1st_arg type of 1st argument
+ * @param value_1st_arg value of 1st argument
+ * @param type_2nd_arg type of 2nd argument
+ * @param value_2nd_arg value of 2nd argument
+ * ... etc
+ * @function emit_signal
+ */
+static int
+luaA_dbus_emit_signal(lua_State *L)
+{
+    const char *bus_name = luaL_checkstring(L, 1);
+    const char *path = luaL_checkstring(L, 2);
+    const char *itface = luaL_checkstring(L, 3);
+    const char *name = luaL_checkstring(L, 4);
+    DBusConnection *dbus_connection = a_dbus_bus_getbyname(bus_name);
+    DBusMessage* msg = dbus_message_new_signal(path, itface, name);
+    if (msg == NULL) {
+        luaA_warn(L, "your D-Bus signal emitting method error'd");
+        return 0;
+    }
+
+    DBusMessageIter iter;
+    dbus_message_iter_init_append(msg, &iter);
+    int top = lua_gettop(L);
+    int nargs = top - 4;
+
+    if(nargs % 2 != 0)
+    {
+        luaA_warn(L, "your D-Bus signal emitting method has wrong number of arguments");
+        dbus_message_unref(msg);
+        lua_pushboolean(L, 0);
+        return 1;
+    }
+    for(int i = 5; i < top; i += 2) {
+        if(!a_dbus_convert_value(L, i, &iter))
+        {
+            luaA_warn(L, "your D-Bus signal emitting method has bad argument type");
+            dbus_message_unref(msg);
+            lua_pushboolean(L, 0);
+            return 1;
+        }
+    }
+    dbus_connection_send(dbus_connection, msg, NULL);
+    dbus_message_unref(msg);
+    dbus_connection_flush(dbus_connection);
+    lua_pushboolean(L, 1);
+    return 1;
+}
+
 const struct luaL_Reg awesome_dbus_lib[] =
 {
     { "request_name", luaA_dbus_request_name },
@@ -797,6 +857,9 @@ const struct luaL_Reg awesome_dbus_lib[] =
     { "remove_match", luaA_dbus_remove_match },
     { "connect_signal", luaA_dbus_connect_signal },
     { "disconnect_signal", luaA_dbus_disconnect_signal },
+    { "emit_signal", luaA_dbus_emit_signal },
+    { "__index", luaA_default_index },
+    { "__newindex", luaA_default_newindex },
     { NULL, NULL }
 };
 
