@@ -19,76 +19,18 @@
  *
  */
 
-#include <cairo-xcb.h>
-
 #include "config.h"
+#include "draw.h"
+#include "globalconf.h"
 
 #include <langinfo.h>
-#include <iconv.h>
 #include <errno.h>
 #include <ctype.h>
 #include <math.h>
 
 #include <gdk-pixbuf/gdk-pixbuf.h>
-
-#include "globalconf.h"
-#include "screen.h"
-
-#include "common/xutil.h"
-
-/** Convert text from any charset to UTF-8 using iconv.
- * \param iso The ISO string to convert.
- * \param len The string size.
- * \param dest The destination pointer. Memory will be allocated, up to you to
- * free, like any char *.
- * \param dlen The destination length, can be NULL.
- * \return True if conversion was done.
- */
-bool
-draw_iso2utf8(const char *iso, size_t len, char **dest, ssize_t *dlen)
-{
-    static iconv_t iso2utf8 = (iconv_t) -1;
-    static int8_t dont_need_convert = -1;
-
-    if(dont_need_convert == -1)
-        dont_need_convert = A_STREQ(nl_langinfo(CODESET), "UTF-8");
-
-    if(!len || dont_need_convert)
-        return false;
-
-    if(iso2utf8 == (iconv_t) -1)
-    {
-        iso2utf8 = iconv_open("UTF-8", nl_langinfo(CODESET));
-        if(iso2utf8 == (iconv_t) -1)
-        {
-            if(errno == EINVAL)
-                warn("unable to convert text from %s to UTF-8, not available",
-                     nl_langinfo(CODESET));
-            else
-                warn("unable to convert text: %s", strerror(errno));
-
-            return false;
-        }
-    }
-
-    size_t orig_utf8len, utf8len;
-    char *utf8;
-
-    orig_utf8len = utf8len = 2 * len + 1;
-    utf8 = *dest = p_new(char, utf8len);
-
-    if(iconv(iso2utf8, (char **) &iso, &len, &utf8, &utf8len) == (size_t) -1)
-    {
-        warn("text conversion failed: %s", strerror(errno));
-        p_delete(dest);
-        return false;
-    }
-
-    if(dlen)
-        *dlen = orig_utf8len - utf8len;
-
-    return true;
-}
+#include <cairo-xcb.h>
+#include <lauxlib.h>
 
 static cairo_user_data_key_t data_key;
 
@@ -99,7 +41,6 @@ free_data(void *data)
 }
 
 /** Create a surface object from this image data.
- * \param L The lua stack.
  * \param width The width of the image.
  * \param height The height of the image
  * \param data The image's data in ARGB format, will be copied by this function.
@@ -192,6 +133,18 @@ draw_surface_from_pixbuf(GdkPixbuf *buf)
     return surface;
 }
 
+static void
+get_surface_size(cairo_surface_t *surface, int *width, int *height)
+{
+    double x1, y1, x2, y2;
+    cairo_t *cr = cairo_create(surface);
+
+    cairo_clip_extents(cr, &x1, &y1, &x2, &y2);
+    cairo_destroy(cr);
+    *width = x2 - x1;
+    *height = y2 - y1;
+}
+
 /** Duplicate the specified image surface.
  * \param surface The surface to copy
  * \return A pointer to a new cairo image surface.
@@ -199,10 +152,15 @@ draw_surface_from_pixbuf(GdkPixbuf *buf)
 cairo_surface_t *
 draw_dup_image_surface(cairo_surface_t *surface)
 {
-    cairo_surface_t *res = cairo_image_surface_create(
-            cairo_image_surface_get_format(surface),
-            cairo_image_surface_get_width(surface),
-            cairo_image_surface_get_height(surface));
+    cairo_surface_t *res;
+    int width, height;
+
+    get_surface_size(surface, &width, &height);
+#if CAIRO_VERSION_MAJOR == 1 && CAIRO_VERSION_MINOR > 12
+    res = cairo_surface_create_similar_image(surface, CAIRO_FORMAT_ARGB32, width, height);
+#else
+    res = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
+#endif
 
     cairo_t *cr = cairo_create(res);
     cairo_set_source_surface(cr, surface, 0, 0);
@@ -216,30 +174,25 @@ draw_dup_image_surface(cairo_surface_t *surface)
 /** Load the specified path into a cairo surface
  * \param L Lua state
  * \param path file to load
+ * \param error A place to store an error message, if needed
  * \return A cairo image surface or NULL on error.
  */
 cairo_surface_t *
-draw_load_image(lua_State *L, const char *path)
+draw_load_image(lua_State *L, const char *path, GError **error)
 {
-    GError *error = NULL;
     cairo_surface_t *ret;
-    GdkPixbuf *buf = gdk_pixbuf_new_from_file(path, &error);
+    GdkPixbuf *buf = gdk_pixbuf_new_from_file(path, error);
 
-    if (!buf) {
-        luaL_where(L, 1);
-        lua_pushstring(L, error->message);
-        lua_concat(L, 2);
-        g_error_free(error);
-        lua_error(L);
+    if (!buf)
+        /* error was set above */
         return NULL;
-    }
 
     ret = draw_surface_from_pixbuf(buf);
     g_object_unref(buf);
     return ret;
 }
 
-xcb_visualtype_t *draw_default_visual(const xcb_screen_t *s)
+xcb_visualtype_t *draw_find_visual(const xcb_screen_t *s, xcb_visualid_t visual)
 {
     xcb_depth_iterator_t depth_iter = xcb_screen_allowed_depths_iterator(s);
 
@@ -247,12 +200,17 @@ xcb_visualtype_t *draw_default_visual(const xcb_screen_t *s)
         for(; depth_iter.rem; xcb_depth_next (&depth_iter))
             for(xcb_visualtype_iterator_t visual_iter = xcb_depth_visuals_iterator(depth_iter.data);
                 visual_iter.rem; xcb_visualtype_next (&visual_iter))
-                if(s->root_visual == visual_iter.data->visual_id)
+                if(visual == visual_iter.data->visual_id)
                     return visual_iter.data;
 
     return NULL;
 }
 
+xcb_visualtype_t *draw_default_visual(const xcb_screen_t *s)
+{
+    return draw_find_visual(s, s->root_visual);
+}
+
 xcb_visualtype_t *draw_argb_visual(const xcb_screen_t *s)
 {
     xcb_depth_iterator_t depth_iter = xcb_screen_allowed_depths_iterator(s);
@@ -281,4 +239,19 @@ uint8_t draw_visual_depth(const xcb_screen_t *s, xcb_visualid_t vis)
     fatal("Could not find a visual's depth");
 }
 
+void draw_test_cairo_xcb(void)
+{
+    xcb_pixmap_t pixmap = xcb_generate_id(globalconf.connection);
+    xcb_create_pixmap(globalconf.connection, globalconf.default_depth, pixmap,
+                      globalconf.screen->root, 1, 1);
+    cairo_surface_t *surface = cairo_xcb_surface_create(globalconf.connection,
+                                          pixmap, globalconf.visual, 1, 1);
+    if(cairo_surface_status(surface) != CAIRO_STATUS_SUCCESS)
+        fatal("Could not set up display: got cairo surface with status %s",
+                cairo_status_to_string(cairo_surface_status(surface)));
+    cairo_surface_finish(surface);
+    cairo_surface_destroy(surface);
+    xcb_free_pixmap(globalconf.connection, pixmap);
+}
+
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
