@@ -25,7 +25,9 @@
 #include <xcb/xcb.h>
 #include <cairo.h>
 #include <lua.h>
+#include <glib.h> /* for GError */
 
+#include "common/array.h"
 #include "common/util.h"
 
 typedef struct area_t area_t;
@@ -42,37 +44,26 @@ struct area_t
 #define AREA_TOP(a)     ((a).y)
 #define AREA_RIGHT(a)   ((a).x + (a).width)
 #define AREA_BOTTOM(a)    ((a).y + (a).height)
+#define AREA_EQUAL(a, b) ((a).x == (b).x && (a).y == (b).y && \
+        (a).width == (b).width && (a).height == (b).height)
 
-bool draw_iso2utf8(const char *, size_t, char **, ssize_t *);
-
-/** Convert a string to UTF-8.
- * \param str The string to convert.
- * \param len The string length.
- * \param dest The destination string that will be allocated.
- * \param dlen The destination string length allocated, can be NULL.
- * \return True if the conversion happened, false otherwise. In both case, dest
- * and dlen will have value and dest have to be free().
- */
-static inline bool
-a_iso2utf8(const char *str, ssize_t len, char **dest, ssize_t *dlen)
+static inline void
+cairo_surface_array_destroy_surface(cairo_surface_t **s)
 {
-    if(draw_iso2utf8(str, len, dest, dlen))
-        return true;
-
-    *dest = a_strdup(str);
-    if(dlen)
-        *dlen = len;
-
-    return false;
+    cairo_surface_destroy(*s);
 }
+DO_ARRAY(cairo_surface_t *, cairo_surface, cairo_surface_array_destroy_surface)
 
 cairo_surface_t *draw_surface_from_data(int width, int height, uint32_t *data);
 cairo_surface_t *draw_dup_image_surface(cairo_surface_t *surface);
-cairo_surface_t *draw_load_image(lua_State *L, const char *path);
+cairo_surface_t *draw_load_image(lua_State *L, const char *path, GError **error);
 
+xcb_visualtype_t *draw_find_visual(const xcb_screen_t *s, xcb_visualid_t visual);
 xcb_visualtype_t *draw_default_visual(const xcb_screen_t *s);
 xcb_visualtype_t *draw_argb_visual(const xcb_screen_t *s);
 uint8_t draw_visual_depth(const xcb_screen_t *s, xcb_visualid_t vis);
 
+void draw_test_cairo_xcb(void);
+
 #endif
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
