@@ -19,13 +19,65 @@
  *
  */
 
+/** awesome mouse API.
+ *
+ * The mouse buttons are represented as index. The common ones are:
+ *
+ * ![Client geometry](../images/mouse.svg)
+ *
+ * It is possible to be notified of mouse events by connecting to various
+ * `client`, `widget`s and `wibox` signals:
+ *
+ *  * `mouse::enter`
+ *  * `mouse::leave`
+ *  * `mouse::press`
+ *  * `mouse::release`
+ *  * `mouse::move`
+ *
+ * It is also possible to add generic mouse button callbacks for `client`s,
+ * `wiboxe`s and the `root` window. Those are set in the default `rc.lua` as such:
+ *
+ * **root**:
+ *
+ *    root.buttons(awful.util.table.join(
+ *        awful.button({ }, 3, function () mymainmenu:toggle() end),
+ *        awful.button({ }, 4, awful.tag.viewnext),
+ *        awful.button({ }, 5, awful.tag.viewprev)
+ *    ))
+ *
+ * **client**:
+ *
+ *    clientbuttons = awful.util.table.join(
+ *        awful.button({ }, 1, function (c) client.focus = c; c:raise() end),
+ *        awful.button({ modkey }, 1, awful.mouse.client.move),
+ *        awful.button({ modkey }, 3, awful.mouse.client.resize)
+ *    )
+ *
+ * See also `mousegrabber`
+ *
+ * @author Julien Danjou &lt;julien@danjou.info&gt;
+ * @copyright 2008-2009 Julien Danjou
+ * @module mouse
+ */
+
 #include "mouse.h"
-#include "screen.h"
-#include "objects/client.h"
-#include "globalconf.h"
-#include "objects/drawin.h"
-#include "luaa.h"
+#include "math.h"
+#include "common/util.h"
 #include "common/xutil.h"
+#include "common/luaclass.h"
+#include "globalconf.h"
+#include "objects/client.h"
+#include "objects/drawin.h"
+#include "objects/screen.h"
+
+static int miss_index_handler    = LUA_REFNIL;
+static int miss_newindex_handler = LUA_REFNIL;
+
+/**
+ * The `screen` under the cursor
+ * @property screen
+ * @param screen
+ */
 
 /** Get the pointer position.
  * \param window The window to get position on.
@@ -84,35 +136,80 @@ mouse_query_pointer_root(int16_t *x, int16_t *y, xcb_window_t *child, uint16_t *
  * \param y Y-coordinate inside window.
  */
 static inline void
-mouse_warp_pointer(xcb_window_t window, int x, int y)
+mouse_warp_pointer(xcb_window_t window, int16_t x, int16_t y)
 {
     xcb_warp_pointer(globalconf.connection, XCB_NONE, window,
                      0, 0, 0, 0, x, y);
 }
 
+/**
+ * Allow the a Lua handler to be implemented for custom properties and
+ * functions.
+ * \param L A lua state
+ * \param handler A function on the LUA_REGISTRYINDEX
+ */
+static int
+luaA_mouse_call_handler(lua_State *L, int handler)
+{
+    int nargs = lua_gettop(L);
+
+    /* Push error handling function and move it before args */
+    lua_pushcfunction(L, luaA_dofunction_error);
+    lua_insert(L, - nargs - 1);
+    int error_func_pos = 1;
+
+    /* push function and move it before args */
+    lua_rawgeti(L, LUA_REGISTRYINDEX, handler);
+    lua_insert(L, - nargs - 1);
+
+    if(lua_pcall(L, nargs, LUA_MULTRET, error_func_pos))
+    {
+        warn("%s", lua_tostring(L, -1));
+        /* Remove error function and error string */
+        lua_pop(L, 2);
+        return 0;
+    }
+    /* Remove error function */
+    lua_remove(L, error_func_pos);
+
+    return lua_gettop(L);
+}
+
 /** Mouse library.
  * \param L The Lua VM state.
  * \return The number of elements pushed on stack.
  * \luastack
  * \lfield coords Mouse coordinates.
- * \lfield screen Mouse screen number.
+ * \lfield screen Mouse screen.
  */
 static int
 luaA_mouse_index(lua_State *L)
 {
     const char *attr = luaL_checkstring(L, 2);
     int16_t mouse_x, mouse_y;
-    screen_t *screen;
 
     /* attr is not "screen"?! */
-    if (A_STRNEQ(attr, "screen"))
-        return 0;
+    if (A_STRNEQ(attr, "screen")) {
+        if (miss_index_handler != LUA_REFNIL) {
+            return luaA_mouse_call_handler(L, miss_index_handler);
+        }
+        else
+            return luaA_default_index(L);
+    }
 
     if (!mouse_query_pointer_root(&mouse_x, &mouse_y, NULL, NULL))
-        return 0;
+    {
+        /* Nothing ever handles mouse.screen being nil. Lying is better than
+         * having lots of lua errors in this case.
+         */
+        if (globalconf.focus.client)
+            luaA_object_push(L, globalconf.focus.client->screen);
+        else
+            luaA_object_push(L, screen_get_primary());
+        return 1;
+    }
 
-    screen = screen_getbycoord(mouse_x, mouse_y);
-    lua_pushnumber(L, screen_array_indexof(&globalconf.screens, screen) + 1);
+    luaA_object_push(L, screen_getbycoord(mouse_x, mouse_y));
     return 1;
 }
 
@@ -124,19 +221,19 @@ static int
 luaA_mouse_newindex(lua_State *L)
 {
     const char *attr = luaL_checkstring(L, 2);
-    int x, y = 0;
-    int screen;
+    screen_t *screen;
 
-    if (A_STRNEQ(attr, "screen"))
-        return 0;
+    if (A_STRNEQ(attr, "screen")) {
+        /* Call the lua mouse property handler */
+        if (miss_newindex_handler != LUA_REFNIL) {
+            return luaA_mouse_call_handler(L, miss_newindex_handler);
+        }
+        else
+            return luaA_default_newindex(L);
+    }
 
-    screen = luaL_checknumber(L, 3) - 1;
-    luaA_checkscreen(screen);
-
-    x = globalconf.screens.tab[screen].geometry.x;
-    y = globalconf.screens.tab[screen].geometry.y;
-
-    mouse_warp_pointer(globalconf.screen->root, x, y);
+    screen = luaA_checkscreen(L, 3);
+    mouse_warp_pointer(globalconf.screen->root, screen->geometry.x, screen->geometry.y);
     return 0;
 }
 
@@ -150,9 +247,9 @@ int
 luaA_mouse_pushstatus(lua_State *L, int x, int y, uint16_t mask)
 {
     lua_createtable(L, 0, 2);
-    lua_pushnumber(L, x);
+    lua_pushinteger(L, x);
     lua_setfield(L, -2, "x");
-    lua_pushnumber(L, y);
+    lua_pushinteger(L, y);
     lua_setfield(L, -2, "y");
 
     lua_createtable(L, 5, 0);
@@ -171,10 +268,7 @@ luaA_mouse_pushstatus(lua_State *L, int x, int y, uint16_t mask)
     return 1;
 }
 
-/** Get or set the mouse coords.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- */
+/* documented in lib/awful/mouse/init.lua */
 static int
 luaA_mouse_coords(lua_State *L)
 {
@@ -190,8 +284,8 @@ luaA_mouse_coords(lua_State *L)
         if(!mouse_query_pointer_root(&mouse_x, &mouse_y, NULL, &mask))
             return 0;
 
-        x = luaA_getopt_number(L, 1, "x", mouse_x);
-        y = luaA_getopt_number(L, 1, "y", mouse_y);
+        x = round(luaA_getopt_number_range(L, 1, "x", mouse_x, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
+        y = round(luaA_getopt_number_range(L, 1, "y", mouse_y, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
 
         if(ignore_enter_notify)
             client_ignore_enterleave_events();
@@ -210,11 +304,10 @@ luaA_mouse_coords(lua_State *L)
     return luaA_mouse_pushstatus(L, mouse_x, mouse_y, mask);
 }
 
-/** Get the client which is under the pointer.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- * \luastack
- * \lreturn A client or nil.
+/** Get the client or any object which is under the pointer.
+ *
+ * @treturn client.object|nil A client or nil.
+ * @function object_under_pointer
  */
 static int
 luaA_mouse_object_under_pointer(lua_State *L)
@@ -232,17 +325,37 @@ luaA_mouse_object_under_pointer(lua_State *L)
         return luaA_object_push(L, drawin);
 
     if((client = client_getbyframewin(child)))
-        return luaA_object_push(globalconf.L, client);
+        return luaA_object_push(L, client);
 
     return 0;
 }
 
+/**
+ * Add a custom property handler (getter).
+ */
+static int
+luaA_mouse_set_index_miss_handler(lua_State *L)
+{
+    return luaA_registerfct(L, 1, &miss_index_handler);
+}
+
+/**
+ * Add a custom property handler (setter).
+ */
+static int
+luaA_mouse_set_newindex_miss_handler(lua_State *L)
+{
+    return luaA_registerfct(L, 1, &miss_newindex_handler);
+}
+
 const struct luaL_Reg awesome_mouse_methods[] =
 {
     { "__index", luaA_mouse_index },
     { "__newindex", luaA_mouse_newindex },
     { "coords", luaA_mouse_coords },
     { "object_under_pointer", luaA_mouse_object_under_pointer },
+    { "set_index_miss_handler", luaA_mouse_set_index_miss_handler},
+    { "set_newindex_miss_handler", luaA_mouse_set_newindex_miss_handler},
     { NULL, NULL }
 };
 const struct luaL_Reg awesome_mouse_meta[] =
