@@ -20,20 +20,130 @@
  *
  */
 
-#include "screen.h"
+/** awesome drawin API
+ *
+ * Furthermore to the classes described here, one can also use signals as
+ * described in @{signals} and X properties as described in @{xproperties}.
+ *
+ * @author Julien Danjou &lt;julien@danjou.info&gt;
+ * @copyright 2008-2009 Julien Danjou
+ * @classmod drawin
+ */
+
 #include "drawin.h"
-#include "objects/client.h"
-#include "screen.h"
-#include "xwindow.h"
-#include "luaa.h"
-#include "ewmh.h"
-#include "systray.h"
+#include "common/atoms.h"
 #include "common/xcursor.h"
 #include "common/xutil.h"
+#include "event.h"
+#include "ewmh.h"
+#include "objects/client.h"
+#include "objects/screen.h"
+#include "systray.h"
+#include "xwindow.h"
+
+#include "math.h"
 
 #include <cairo-xcb.h>
 #include <xcb/shape.h>
 
+/** Drawin object.
+ *
+ * @field border_width Border width.
+ * @field border_color Border color.
+ * @field ontop On top of other windows.
+ * @field cursor The mouse cursor.
+ * @field visible Visibility.
+ * @field opacity The opacity of the drawin, between 0 and 1.
+ * @field type The window type (desktop, normal, dock, …).
+ * @field x The x coordinates.
+ * @field y The y coordinates.
+ * @field width The width of the drawin.
+ * @field height The height of the drawin.
+ * @field drawable The drawin's drawable.
+ * @field window The X window id.
+ * @field shape_bounding The drawin's bounding shape as a (native) cairo surface.
+ * @field shape_clip The drawin's clip shape as a (native) cairo surface.
+ * @field shape_input The drawin's input shape as a (native) cairo surface.
+ * @table drawin
+ */
+
+/**
+ * @signal property::geometry
+ */
+
+/**
+ * @signal property::shape_bounding
+ */
+
+/**
+ * @signal property::shape_clip
+ */
+
+/**
+ * @signal property::shape_input
+ */
+
+/**
+ * @signal property::border_width
+ */
+
+/**
+ * @signal property::cursor
+ */
+
+/**
+ * @signal property::height
+ */
+
+/**
+ * @signal property::ontop
+ */
+
+/**
+ * @signal property::visible
+ */
+
+/**
+ * @signal property::width
+ */
+
+/**
+ * @signal property::x
+ */
+
+/**
+ * @signal property::y
+ */
+
+/** Get or set mouse buttons bindings to a drawin.
+ *
+ * @param buttons_table A table of buttons objects, or nothing.
+ * @function buttons
+ */
+
+/** Get or set drawin struts.
+ *
+ * @param strut A table with new strut, or nothing
+ * @return The drawin strut in a table.
+ * @function struts
+ */
+
+/** Get the number of instances.
+ *
+ * @return The number of drawin objects alive.
+ * @function instances
+ */
+
+/** Set a __index metamethod for all drawin instances.
+ * @tparam function cb The meta-method
+ * @function set_index_miss_handler
+ */
+
+/** Set a __newindex metamethod for all drawin instances.
+ * @tparam function cb The meta-method
+ * @function set_newindex_miss_handler
+ */
+
 LUA_OBJECT_FUNCS(drawin_class, drawin_t, drawin)
 
 /** Kick out systray windows.
@@ -56,98 +166,36 @@ drawin_systray_kickout(drawin_t *w)
     }
 }
 
+void
+luaA_drawin_systray_kickout(lua_State *L)
+{
+    drawin_systray_kickout(luaA_checkudata(L, 1, &drawin_class));
+}
+
 static void
 drawin_wipe(drawin_t *w)
 {
     /* The drawin must already be unmapped, else it
      * couldn't be garbage collected -> no unmap needed */
     p_delete(&w->cursor);
-    drawable_unset_surface(w->drawable);
     if(w->window)
     {
-        /* Activate BMA */
-        client_ignore_enterleave_events();
         /* Make sure we don't accidentally kill the systray window */
         drawin_systray_kickout(w);
         xcb_destroy_window(globalconf.connection, w->window);
-        /* Deactivate BMA */
-        client_restore_enterleave_events();
         w->window = XCB_NONE;
     }
-    if(w->pixmap)
-    {
-        xcb_free_pixmap(globalconf.connection, w->pixmap);
-        w->pixmap = XCB_NONE;
-    }
-    luaA_object_unref_item(globalconf.L, -1, w->drawable);
+    /* No unref needed because we are being garbage collected */
     w->drawable = NULL;
 }
 
-void
-drawin_unref_simplified(drawin_t **item)
-{
-    luaA_object_unref(globalconf.L, *item);
-}
-
 static void
-drawin_update_drawing(drawin_t *w, int widx)
+drawin_update_drawing(lua_State *L, int widx)
 {
-    /* If this drawin isn't visible, we don't need an up-to-date cairo surface
-     * for it. (drawin_map() will later make sure we are called again) */
-    if(!w->visible)
-        return;
-    /* Clean up old stuff */
-    luaA_object_push_item(globalconf.L, widx, w->drawable);
-    drawable_unset_surface(w->drawable);
-    if(w->pixmap)
-        xcb_free_pixmap(globalconf.connection, w->pixmap);
-
-    /* Create a pixmap */
-    xcb_screen_t *s = globalconf.screen;
-    w->pixmap = xcb_generate_id(globalconf.connection);
-    xcb_create_pixmap(globalconf.connection, globalconf.default_depth, w->pixmap, s->root,
-                      w->geometry.width, w->geometry.height);
-    /* and create a surface for that pixmap */
-    cairo_surface_t *surface = cairo_xcb_surface_create(globalconf.connection,
-                                                        w->pixmap, globalconf.visual,
-                                                        w->geometry.width, w->geometry.height);
-    drawable_set_surface(w->drawable, -1, surface, w->geometry);
-    lua_pop(globalconf.L, 1);
-}
-
-/** Initialize a drawin.
- * \param w The drawin to initialize.
- */
-static void
-drawin_init(drawin_t *w)
-{
-    xcb_screen_t *s = globalconf.screen;
-
-    w->window = xcb_generate_id(globalconf.connection);
-    xcb_create_window(globalconf.connection, globalconf.default_depth, w->window, s->root,
-                      w->geometry.x, w->geometry.y,
-                      w->geometry.width, w->geometry.height,
-                      w->border_width, XCB_COPY_FROM_PARENT, globalconf.visual->visual_id,
-                      XCB_CW_BORDER_PIXEL | XCB_CW_BIT_GRAVITY
-                      | XCB_CW_OVERRIDE_REDIRECT | XCB_CW_EVENT_MASK | XCB_CW_COLORMAP
-                      | XCB_CW_CURSOR,
-                      (const uint32_t [])
-                      {
-                          w->border_color.pixel,
-                          XCB_GRAVITY_NORTH_WEST,
-                          1,
-                          XCB_EVENT_MASK_SUBSTRUCTURE_REDIRECT
-                          | XCB_EVENT_MASK_SUBSTRUCTURE_NOTIFY | XCB_EVENT_MASK_ENTER_WINDOW
-                          | XCB_EVENT_MASK_LEAVE_WINDOW | XCB_EVENT_MASK_STRUCTURE_NOTIFY
-                          | XCB_EVENT_MASK_POINTER_MOTION | XCB_EVENT_MASK_EXPOSURE
-                          | XCB_EVENT_MASK_PROPERTY_CHANGE,
-                          globalconf.default_cmap,
-                          xcursor_new(globalconf.display, xcursor_font_fromstr(w->cursor))
-                      });
-
-    /* Set the right properties */
-    ewmh_update_window_type(w->window, window_translate_type(w->type));
-    ewmh_update_strut(w->window, &w->strut);
+    drawin_t *w = luaA_checkudata(L, widx, &drawin_class);
+    luaA_object_push_item(L, widx, w->drawable);
+    drawable_set_geometry(L, -1, w->geometry);
+    lua_pop(L, 1);
 }
 
 /** Refresh the window content by copying its pixmap data to its window.
@@ -159,6 +207,37 @@ drawin_refresh_pixmap(drawin_t *w)
     drawin_refresh_pixmap_partial(w, 0, 0, w->geometry.width, w->geometry.height);
 }
 
+static void
+drawin_apply_moveresize(drawin_t *w)
+{
+    if (!w->geometry_dirty)
+        return;
+
+    w->geometry_dirty = false;
+    client_ignore_enterleave_events();
+    xcb_configure_window(globalconf.connection, w->window,
+                         XCB_CONFIG_WINDOW_X | XCB_CONFIG_WINDOW_Y
+                         | XCB_CONFIG_WINDOW_WIDTH | XCB_CONFIG_WINDOW_HEIGHT,
+                         (const uint32_t [])
+                         {
+                             w->geometry.x,
+                             w->geometry.y,
+                             w->geometry.width,
+                             w->geometry.height
+                         });
+    client_restore_enterleave_events();
+}
+
+void
+drawin_refresh(void)
+{
+    foreach(item, globalconf.drawins)
+    {
+        drawin_apply_moveresize(*item);
+        window_border_refresh((window_t *) *item);
+    }
+}
+
 /** Move and/or resize a drawin
  * \param L The Lua VM state.
  * \param udx The index of the drawin.
@@ -168,59 +247,35 @@ static void
 drawin_moveresize(lua_State *L, int udx, area_t geometry)
 {
     drawin_t *w = luaA_checkudata(L, udx, &drawin_class);
-    int number_of_vals = 0;
-    uint32_t moveresize_win_vals[4], mask_vals = 0;
+    area_t old_geometry = w->geometry;
 
-    if(w->geometry.x != geometry.x)
-    {
-        w->geometry.x = moveresize_win_vals[number_of_vals++] = geometry.x;
-        mask_vals |= XCB_CONFIG_WINDOW_X;
-    }
+    w->geometry = geometry;
+    if(w->geometry.width <= 0)
+        w->geometry.width = old_geometry.width;
+    if(w->geometry.height <= 0)
+        w->geometry.height = old_geometry.height;
 
-    if(w->geometry.y != geometry.y)
-    {
-        w->geometry.y = moveresize_win_vals[number_of_vals++] = geometry.y;
-        mask_vals |= XCB_CONFIG_WINDOW_Y;
-    }
+    w->geometry_dirty = true;
+    drawin_update_drawing(L, udx);
 
-    if(geometry.width > 0 && w->geometry.width != geometry.width)
-    {
-        w->geometry.width = moveresize_win_vals[number_of_vals++] = geometry.width;
-        mask_vals |= XCB_CONFIG_WINDOW_WIDTH;
-    }
-
-    if(geometry.height > 0 && w->geometry.height != geometry.height)
-    {
-        w->geometry.height = moveresize_win_vals[number_of_vals++] = geometry.height;
-        mask_vals |= XCB_CONFIG_WINDOW_HEIGHT;
-    }
-
-    if(mask_vals & (XCB_CONFIG_WINDOW_WIDTH | XCB_CONFIG_WINDOW_HEIGHT))
-        drawin_update_drawing(w, udx);
-    else {
-        /* We still have to set x/y */
-        luaA_object_push_item(L, udx, w->drawable);
-        drawable_set_geometry(w->drawable, -1, w->geometry);
-        lua_pop(L, 1);
-    }
-
-    /* Activate BMA */
-    client_ignore_enterleave_events();
-
-    if(mask_vals)
-        xcb_configure_window(globalconf.connection, w->window, mask_vals, moveresize_win_vals);
-
-    /* Deactivate BMA */
-    client_restore_enterleave_events();
-
-    if(mask_vals & XCB_CONFIG_WINDOW_X)
+    if (!AREA_EQUAL(old_geometry, w->geometry))
+        luaA_object_emit_signal(L, udx, "property::geometry", 0);
+    if (old_geometry.x != w->geometry.x)
         luaA_object_emit_signal(L, udx, "property::x", 0);
-    if(mask_vals & XCB_CONFIG_WINDOW_Y)
+    if (old_geometry.y != w->geometry.y)
         luaA_object_emit_signal(L, udx, "property::y", 0);
-    if(mask_vals & XCB_CONFIG_WINDOW_WIDTH)
+    if (old_geometry.width != w->geometry.width)
         luaA_object_emit_signal(L, udx, "property::width", 0);
-    if(mask_vals & XCB_CONFIG_WINDOW_HEIGHT)
+    if (old_geometry.height != w->geometry.height)
         luaA_object_emit_signal(L, udx, "property::height", 0);
+
+    screen_t *old_screen = screen_getbycoord(old_geometry.x, old_geometry.y);
+    screen_t *new_screen = screen_getbycoord(w->geometry.x, w->geometry.y);
+    if (old_screen != new_screen && strut_has_value(&w->strut))
+    {
+        screen_update_workarea(old_screen);
+        screen_update_workarea(new_screen);
+    }
 }
 
 /** Refresh the window content by copying its pixmap data to its window.
@@ -235,30 +290,38 @@ drawin_refresh_pixmap_partial(drawin_t *drawin,
                               int16_t x, int16_t y,
                               uint16_t w, uint16_t h)
 {
-    if (!drawin->visible)
+    if (!drawin->drawable || !drawin->drawable->pixmap || !drawin->drawable->refreshed)
         return;
 
+    /* Make sure it really has the size it should have */
+    drawin_apply_moveresize(drawin);
+
     /* Make cairo do all pending drawing */
     cairo_surface_flush(drawin->drawable->surface);
-    xcb_copy_area(globalconf.connection, drawin->pixmap,
+    xcb_copy_area(globalconf.connection, drawin->drawable->pixmap,
                   drawin->window, globalconf.gc, x, y, x, y,
                   w, h);
 }
 
 static void
-drawin_map(drawin_t *drawin, int widx)
+drawin_map(lua_State *L, int widx)
 {
+    drawin_t *drawin = luaA_checkudata(L, widx, &drawin_class);
+    /* Apply any pending changes */
+    drawin_apply_moveresize(drawin);
     /* Activate BMA */
     client_ignore_enterleave_events();
     /* Map the drawin */
     xcb_map_window(globalconf.connection, drawin->window);
-    drawin_update_drawing(drawin, widx);
     /* Deactivate BMA */
     client_restore_enterleave_events();
     /* Stack this drawin correctly */
     stack_windows();
     /* Add it to the list of visible drawins */
     drawin_array_append(&globalconf.drawins, drawin);
+    /* Make sure it has a surface */
+    if(drawin->drawable->surface == NULL)
+        drawin_update_drawing(L, widx);
 }
 
 static void
@@ -301,11 +364,11 @@ drawin_set_visible(lua_State *L, int udx, bool v)
 
         if(drawin->visible)
         {
-            drawin_map(drawin, udx);
+            drawin_map(L, udx);
             /* duplicate drawin */
             lua_pushvalue(L, udx);
             /* ref it */
-            luaA_object_ref_class(globalconf.L, -1, &drawin_class);
+            luaA_object_ref_class(L, -1, &drawin_class);
         }
         else
         {
@@ -316,15 +379,14 @@ drawin_set_visible(lua_State *L, int udx, bool v)
             /* Active BMA */
             client_restore_enterleave_events();
             /* unref it */
-            luaA_object_unref(globalconf.L, drawin);
+            luaA_object_unref(L, drawin);
         }
 
         luaA_object_emit_signal(L, udx, "property::visible", 0);
         if(strut_has_value(&drawin->strut))
         {
-            screen_t *screen =
-                screen_getbycoord(drawin->geometry.x, drawin->geometry.y);
-            screen_emit_signal(globalconf.L, screen, "property::workarea", 0);
+            screen_update_workarea(
+                    screen_getbycoord(drawin->geometry.x, drawin->geometry.y));
         }
     }
 }
@@ -332,6 +394,7 @@ drawin_set_visible(lua_State *L, int udx, bool v)
 static drawin_t *
 drawin_allocator(lua_State *L)
 {
+    xcb_screen_t *s = globalconf.screen;
     drawin_t *w = drawin_new(L);
 
     w->visible = false;
@@ -340,12 +403,40 @@ drawin_allocator(lua_State *L)
     w->cursor = a_strdup("left_ptr");
     w->geometry.width = 1;
     w->geometry.height = 1;
+    w->geometry_dirty = false;
     w->type = _NET_WM_WINDOW_TYPE_NORMAL;
 
     drawable_allocator(L, (drawable_refresh_callback *) drawin_refresh_pixmap, w);
     w->drawable = luaA_object_ref_item(L, -2, -1);
 
-    drawin_init(w);
+    w->window = xcb_generate_id(globalconf.connection);
+    xcb_create_window(globalconf.connection, globalconf.default_depth, w->window, s->root,
+                      w->geometry.x, w->geometry.y,
+                      w->geometry.width, w->geometry.height,
+                      w->border_width, XCB_COPY_FROM_PARENT, globalconf.visual->visual_id,
+                      XCB_CW_BORDER_PIXEL | XCB_CW_BIT_GRAVITY
+                      | XCB_CW_OVERRIDE_REDIRECT | XCB_CW_EVENT_MASK | XCB_CW_COLORMAP
+                      | XCB_CW_CURSOR,
+                      (const uint32_t [])
+                      {
+                          w->border_color.pixel,
+                          XCB_GRAVITY_NORTH_WEST,
+                          1,
+                          XCB_EVENT_MASK_SUBSTRUCTURE_REDIRECT
+                          | XCB_EVENT_MASK_SUBSTRUCTURE_NOTIFY | XCB_EVENT_MASK_ENTER_WINDOW
+                          | XCB_EVENT_MASK_LEAVE_WINDOW | XCB_EVENT_MASK_STRUCTURE_NOTIFY
+                          | XCB_EVENT_MASK_POINTER_MOTION | XCB_EVENT_MASK_BUTTON_PRESS
+                          | XCB_EVENT_MASK_BUTTON_RELEASE | XCB_EVENT_MASK_EXPOSURE
+                          | XCB_EVENT_MASK_PROPERTY_CHANGE,
+                          globalconf.default_cmap,
+                          xcursor_new(globalconf.cursor_ctx, xcursor_font_fromstr(w->cursor))
+                      });
+    xwindow_set_class_instance(w->window);
+    xwindow_set_name_static(w->window, "Awesome drawin");
+
+    /* Set the right properties */
+    ewmh_update_window_type(w->window, window_translate_type(w->type));
+    ewmh_update_strut(w->window, &w->strut);
 
     return w;
 }
@@ -362,12 +453,12 @@ luaA_drawin_new(lua_State *L)
     return 1;
 }
 
-/* Set or get the drawin geometry.
- * \param L The Lua VM state.
- * \return The number of elements pushed on stack.
- * \luastack
- * \lparam An optional table with drawin geometry.
- * \lreturn The drawin geometry.
+/** Get or set drawin geometry. That's the same as accessing or setting the x,
+ * y, width or height properties of a drawin.
+ *
+ * @param A table with coordinates to modify.
+ * @return A table with drawin coordinates and geometry.
+ * @function geometry
  */
 static int
 luaA_drawin_geometry(lua_State *L)
@@ -379,10 +470,10 @@ luaA_drawin_geometry(lua_State *L)
         area_t wingeom;
 
         luaA_checktable(L, 2);
-        wingeom.x = luaA_getopt_number(L, 2, "x", drawin->geometry.x);
-        wingeom.y = luaA_getopt_number(L, 2, "y", drawin->geometry.y);
-        wingeom.width = luaA_getopt_number(L, 2, "width", drawin->geometry.width);
-        wingeom.height = luaA_getopt_number(L, 2, "height", drawin->geometry.height);
+        wingeom.x = round(luaA_getopt_number_range(L, 2, "x", drawin->geometry.x, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
+        wingeom.y = round(luaA_getopt_number_range(L, 2, "y", drawin->geometry.y, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
+        wingeom.width = ceil(luaA_getopt_number_range(L, 2, "width", drawin->geometry.width, MIN_X11_SIZE, MAX_X11_SIZE));
+        wingeom.height = ceil(luaA_getopt_number_range(L, 2, "height", drawin->geometry.height, MIN_X11_SIZE, MAX_X11_SIZE));
 
         if(wingeom.width > 0 && wingeom.height > 0)
             drawin_moveresize(L, 1, wingeom);
@@ -399,7 +490,8 @@ LUA_OBJECT_EXPORT_PROPERTY(drawin, drawin_t, visible, lua_pushboolean)
 static int
 luaA_drawin_set_x(lua_State *L, drawin_t *drawin)
 {
-    drawin_moveresize(L, -3, (area_t) { .x = luaL_checknumber(L, -1),
+    int x = round(luaA_checknumber_range(L, -1, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
+    drawin_moveresize(L, -3, (area_t) { .x = x,
                                         .y = drawin->geometry.y,
                                         .width = drawin->geometry.width,
                                         .height = drawin->geometry.height });
@@ -409,15 +501,16 @@ luaA_drawin_set_x(lua_State *L, drawin_t *drawin)
 static int
 luaA_drawin_get_x(lua_State *L, drawin_t *drawin)
 {
-    lua_pushnumber(L, drawin->geometry.x);
+    lua_pushinteger(L, drawin->geometry.x);
     return 1;
 }
 
 static int
 luaA_drawin_set_y(lua_State *L, drawin_t *drawin)
 {
+    int y = round(luaA_checknumber_range(L, -1, MIN_X11_COORDINATE, MAX_X11_COORDINATE));
     drawin_moveresize(L, -3, (area_t) { .x = drawin->geometry.x,
-                                        .y = luaL_checknumber(L, -1),
+                                        .y = y,
                                         .width = drawin->geometry.width,
                                         .height = drawin->geometry.height });
     return 0;
@@ -426,16 +519,14 @@ luaA_drawin_set_y(lua_State *L, drawin_t *drawin)
 static int
 luaA_drawin_get_y(lua_State *L, drawin_t *drawin)
 {
-    lua_pushnumber(L, drawin->geometry.y);
+    lua_pushinteger(L, drawin->geometry.y);
     return 1;
 }
 
 static int
 luaA_drawin_set_width(lua_State *L, drawin_t *drawin)
 {
-    int width = luaL_checknumber(L, -1);
-    if(width <= 0)
-        luaL_error(L, "invalid width");
+    int width = ceil(luaA_checknumber_range(L, -1, MIN_X11_SIZE, MAX_X11_SIZE));
     drawin_moveresize(L, -3, (area_t) { .x = drawin->geometry.x,
                                         .y = drawin->geometry.y,
                                         .width = width,
@@ -446,16 +537,14 @@ luaA_drawin_set_width(lua_State *L, drawin_t *drawin)
 static int
 luaA_drawin_get_width(lua_State *L, drawin_t *drawin)
 {
-    lua_pushnumber(L, drawin->geometry.width);
+    lua_pushinteger(L, drawin->geometry.width);
     return 1;
 }
 
 static int
 luaA_drawin_set_height(lua_State *L, drawin_t *drawin)
 {
-    int height = luaL_checknumber(L, -1);
-    if(height <= 0)
-        luaL_error(L, "invalid height");
+    int height = ceil(luaA_checknumber_range(L, -1, MIN_X11_SIZE, MAX_X11_SIZE));
     drawin_moveresize(L, -3, (area_t) { .x = drawin->geometry.x,
                                        .y = drawin->geometry.y,
                                        .width = drawin->geometry.width,
@@ -466,7 +555,7 @@ luaA_drawin_set_height(lua_State *L, drawin_t *drawin)
 static int
 luaA_drawin_get_height(lua_State *L, drawin_t *drawin)
 {
-    lua_pushnumber(L, drawin->geometry.height);
+    lua_pushinteger(L, drawin->geometry.height);
     return 1;
 }
 
@@ -502,7 +591,7 @@ luaA_drawin_set_cursor(lua_State *L, drawin_t *drawin)
         uint16_t cursor_font = xcursor_font_fromstr(buf);
         if(cursor_font)
         {
-            xcb_cursor_t cursor = xcursor_new(globalconf.display, cursor_font);
+            xcb_cursor_t cursor = xcursor_new(globalconf.cursor_ctx, cursor_font);
             p_delete(&drawin->cursor);
             drawin->cursor = a_strdup(buf);
             xwindow_set_cursor(drawin->window, cursor);
@@ -536,6 +625,22 @@ luaA_drawin_get_drawable(lua_State *L, drawin_t *drawin)
     return 1;
 }
 
+/** Get the drawin's bounding shape.
+ * \param L The Lua VM state.
+ * \param drawin The drawin object.
+ * \return The number of elements pushed on stack.
+ */
+static int
+luaA_drawin_get_shape_bounding(lua_State *L, drawin_t *drawin)
+{
+    cairo_surface_t *surf = xwindow_get_shape(drawin->window, XCB_SHAPE_SK_BOUNDING);
+    if (!surf)
+        return 0;
+    /* lua has to make sure to free the ref or we have a leak */
+    lua_pushlightuserdata(L, surf);
+    return 1;
+}
+
 /** Set the drawin's bounding shape.
  * \param L The Lua VM state.
  * \param drawin The drawin object.
@@ -547,11 +652,34 @@ luaA_drawin_set_shape_bounding(lua_State *L, drawin_t *drawin)
     cairo_surface_t *surf = NULL;
     if(!lua_isnil(L, -1))
         surf = (cairo_surface_t *)lua_touserdata(L, -1);
-    xwindow_set_shape(drawin->window, drawin->geometry.width, drawin->geometry.height,
+
+    /* The drawin might have been resized to a larger size. Apply that. */
+    drawin_apply_moveresize(drawin);
+
+    xwindow_set_shape(drawin->window,
+            drawin->geometry.width + 2*drawin->border_width,
+            drawin->geometry.height + 2*drawin->border_width,
             XCB_SHAPE_SK_BOUNDING, surf, -drawin->border_width);
+    luaA_object_emit_signal(L, -3, "property::shape_bounding", 0);
     return 0;
 }
 
+/** Get the drawin's clip shape.
+ * \param L The Lua VM state.
+ * \param drawin The drawin object.
+ * \return The number of elements pushed on stack.
+ */
+static int
+luaA_drawin_get_shape_clip(lua_State *L, drawin_t *drawin)
+{
+    cairo_surface_t *surf = xwindow_get_shape(drawin->window, XCB_SHAPE_SK_CLIP);
+    if (!surf)
+        return 0;
+    /* lua has to make sure to free the ref or we have a leak */
+    lua_pushlightuserdata(L, surf);
+    return 1;
+}
+
 /** Set the drawin's clip shape.
  * \param L The Lua VM state.
  * \param drawin The drawin object.
@@ -563,8 +691,52 @@ luaA_drawin_set_shape_clip(lua_State *L, drawin_t *drawin)
     cairo_surface_t *surf = NULL;
     if(!lua_isnil(L, -1))
         surf = (cairo_surface_t *)lua_touserdata(L, -1);
+
+    /* The drawin might have been resized to a larger size. Apply that. */
+    drawin_apply_moveresize(drawin);
+
     xwindow_set_shape(drawin->window, drawin->geometry.width, drawin->geometry.height,
             XCB_SHAPE_SK_CLIP, surf, 0);
+    luaA_object_emit_signal(L, -3, "property::shape_clip", 0);
+    return 0;
+}
+
+/** Get the drawin's input shape.
+ * \param L The Lua VM state.
+ * \param drawin The drawin object.
+ * \return The number of elements pushed on stack.
+ */
+static int
+luaA_drawin_get_shape_input(lua_State *L, drawin_t *drawin)
+{
+    cairo_surface_t *surf = xwindow_get_shape(drawin->window, XCB_SHAPE_SK_INPUT);
+    if (!surf)
+        return 0;
+    /* lua has to make sure to free the ref or we have a leak */
+    lua_pushlightuserdata(L, surf);
+    return 1;
+}
+
+/** Set the drawin's input shape.
+ * \param L The Lua VM state.
+ * \param drawin The drawin object.
+ * \return The number of elements pushed on stack.
+ */
+static int
+luaA_drawin_set_shape_input(lua_State *L, drawin_t *drawin)
+{
+    cairo_surface_t *surf = NULL;
+    if(!lua_isnil(L, -1))
+        surf = (cairo_surface_t *)lua_touserdata(L, -1);
+
+    /* The drawin might have been resized to a larger size. Apply that. */
+    drawin_apply_moveresize(drawin);
+
+    xwindow_set_shape(drawin->window,
+            drawin->geometry.width + 2*drawin->border_width,
+            drawin->geometry.height + 2*drawin->border_width,
+            XCB_SHAPE_SK_INPUT, surf, -drawin->border_width);
+    luaA_object_emit_signal(L, -3, "property::shape_input", 0);
     return 0;
 }
 
@@ -630,21 +802,16 @@ drawin_class_setup(lua_State *L)
                             (lua_class_propfunc_t) luaA_window_set_type);
     luaA_class_add_property(&drawin_class, "shape_bounding",
                             (lua_class_propfunc_t) luaA_drawin_set_shape_bounding,
-                            NULL,
+                            (lua_class_propfunc_t) luaA_drawin_get_shape_bounding,
                             (lua_class_propfunc_t) luaA_drawin_set_shape_bounding);
     luaA_class_add_property(&drawin_class, "shape_clip",
                             (lua_class_propfunc_t) luaA_drawin_set_shape_clip,
-                            NULL,
+                            (lua_class_propfunc_t) luaA_drawin_get_shape_clip,
                             (lua_class_propfunc_t) luaA_drawin_set_shape_clip);
-
-    signal_add(&drawin_class.signals, "property::border_width");
-    signal_add(&drawin_class.signals, "property::cursor");
-    signal_add(&drawin_class.signals, "property::height");
-    signal_add(&drawin_class.signals, "property::ontop");
-    signal_add(&drawin_class.signals, "property::visible");
-    signal_add(&drawin_class.signals, "property::width");
-    signal_add(&drawin_class.signals, "property::x");
-    signal_add(&drawin_class.signals, "property::y");
+    luaA_class_add_property(&drawin_class, "shape_input",
+                            (lua_class_propfunc_t) luaA_drawin_set_shape_input,
+                            (lua_class_propfunc_t) luaA_drawin_get_shape_input,
+                            (lua_class_propfunc_t) luaA_drawin_set_shape_input);
 }
 
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
