@@ -22,9 +22,11 @@
 #ifndef AWESOME_OBJECTS_WINDOW_H
 #define AWESOME_OBJECTS_WINDOW_H
 
-#include "strut.h"
-#include "objects/button.h"
+#include "color.h"
 #include "common/luaclass.h"
+#include "objects/button.h"
+#include "strut.h"
+#include "math.h"
 
 /** Windows type */
 typedef enum
@@ -61,6 +63,8 @@ typedef enum
     strut_t strut; \
     /** Button bindings */ \
     button_array_t buttons; \
+    /** Do we have pending border changes? */ \
+    bool border_need_update; \
     /** Border color */ \
     color_t border_color; \
     /** Border width */ \
@@ -82,6 +86,7 @@ void window_class_setup(lua_State *);
 
 void window_set_opacity(lua_State *, int, double);
 void window_set_border_width(lua_State *, int, int);
+void window_border_refresh(window_t *);
 int luaA_window_get_type(lua_State *, window_t *);
 int luaA_window_set_type(lua_State *, window_t *);
 uint32_t window_translate_type(window_type_t);
