@@ -19,17 +19,15 @@
  *
  */
 
-#include <xcb/xcb_atom.h>
-
-#include "screen.h"
 #include "property.h"
-#include "objects/client.h"
-#include "ewmh.h"
-#include "objects/drawin.h"
-#include "xwindow.h"
-#include "luaa.h"
 #include "common/atoms.h"
 #include "common/xutil.h"
+#include "ewmh.h"
+#include "objects/client.h"
+#include "objects/drawin.h"
+#include "xwindow.h"
+
+#include <xcb/xcb_atom.h>
 
 #define HANDLE_TEXT_PROPERTY(funcname, atom, setfunc) \
     xcb_get_property_cookie_t \
@@ -46,11 +44,12 @@
     void \
     property_update_##funcname(client_t *c, xcb_get_property_cookie_t cookie) \
     { \
+        lua_State *L = globalconf_get_lua_State(); \
         xcb_get_property_reply_t * reply = \
                     xcb_get_property_reply(globalconf.connection, cookie, NULL); \
-        luaA_object_push(globalconf.L, c); \
-        setfunc(globalconf.L, -1, xutil_get_text_property_from_reply(reply)); \
-        lua_pop(globalconf.L, 1); \
+        luaA_object_push(L, c); \
+        setfunc(L, -1, xutil_get_text_property_from_reply(reply)); \
+        lua_pop(L, 1); \
         p_delete(&reply); \
     } \
     static int \
@@ -104,6 +103,7 @@ property_get_wm_transient_for(client_t *c)
 void
 property_update_wm_transient_for(client_t *c, xcb_get_property_cookie_t cookie)
 {
+    lua_State *L = globalconf_get_lua_State();
     xcb_window_t trans;
 
     if(!xcb_icccm_get_wm_transient_for_reply(globalconf.connection,
@@ -111,18 +111,22 @@ property_update_wm_transient_for(client_t *c, xcb_get_property_cookie_t cookie)
 					     &trans, NULL))
             return;
 
-    luaA_object_push(globalconf.L, c);
-    client_set_type(globalconf.L, -1, WINDOW_TYPE_DIALOG);
-    client_set_above(globalconf.L, -1, false);
-    client_set_transient_for(globalconf.L, -1, client_getbywin(trans));
-    lua_pop(globalconf.L, 1);
+    c->transient_for_window = trans;
+
+    luaA_object_push(L, c);
+    if (!c->has_NET_WM_WINDOW_TYPE)
+        client_set_type(L, -1, trans == XCB_NONE ? WINDOW_TYPE_NORMAL : WINDOW_TYPE_DIALOG);
+    client_set_above(L, -1, false);
+    lua_pop(L, 1);
+
+    client_find_transient_for(c);
 }
 
 xcb_get_property_cookie_t
 property_get_wm_client_leader(client_t *c)
 {
     return xcb_get_property_unchecked(globalconf.connection, false, c->window,
-                                      WM_CLIENT_LEADER, XCB_ATOM_WINDOW, 1, 32);
+                                      WM_CLIENT_LEADER, XCB_ATOM_WINDOW, 0, 32);
 }
 
 /** Update leader hint of a client.
@@ -156,9 +160,15 @@ property_get_wm_normal_hints(client_t *c)
 void
 property_update_wm_normal_hints(client_t *c, xcb_get_property_cookie_t cookie)
 {
+    lua_State *L = globalconf_get_lua_State();
+
     xcb_icccm_get_wm_normal_hints_reply(globalconf.connection,
 					cookie,
 					&c->size_hints, NULL);
+
+    luaA_object_push(L, c);
+    luaA_object_emit_signal(L, -1, "property::size_hints", 0);
+    lua_pop(L, 1);
 }
 
 xcb_get_property_cookie_t
@@ -174,6 +184,7 @@ property_get_wm_hints(client_t *c)
 void
 property_update_wm_hints(client_t *c, xcb_get_property_cookie_t cookie)
 {
+    lua_State *L = globalconf_get_lua_State();
     xcb_icccm_wm_hints_t wmh;
 
     if(!xcb_icccm_get_wm_hints_reply(globalconf.connection,
@@ -181,16 +192,29 @@ property_update_wm_hints(client_t *c, xcb_get_property_cookie_t cookie)
 				     &wmh, NULL))
         return;
 
-    luaA_object_push(globalconf.L, c);
-    client_set_urgent(globalconf.L, -1, xcb_icccm_wm_hints_get_urgency(&wmh));
+    luaA_object_push(L, c);
+
+    lua_pushboolean(L, xcb_icccm_wm_hints_get_urgency(&wmh));
+    luaA_object_emit_signal(L, -2, "request::urgent", 1);
 
     if(wmh.flags & XCB_ICCCM_WM_HINT_INPUT)
         c->nofocus = !wmh.input;
 
     if(wmh.flags & XCB_ICCCM_WM_HINT_WINDOW_GROUP)
-        client_set_group_window(globalconf.L, -1, wmh.window_group);
+        client_set_group_window(L, -1, wmh.window_group);
 
-    lua_pop(globalconf.L, 1);
+    if(!c->have_ewmh_icon)
+    {
+        if(wmh.flags & XCB_ICCCM_WM_HINT_ICON_PIXMAP)
+        {
+            if(wmh.flags & XCB_ICCCM_WM_HINT_ICON_MASK)
+                client_set_icon_from_pixmaps(c, wmh.icon_pixmap, wmh.icon_mask);
+            else
+                client_set_icon_from_pixmaps(c, wmh.icon_pixmap, XCB_NONE);
+        }
+    }
+
+    lua_pop(L, 1);
 }
 
 xcb_get_property_cookie_t
@@ -206,6 +230,7 @@ property_get_wm_class(client_t *c)
 void
 property_update_wm_class(client_t *c, xcb_get_property_cookie_t cookie)
 {
+    lua_State *L = globalconf_get_lua_State();
     xcb_icccm_get_wm_class_reply_t hint;
 
     if(!xcb_icccm_get_wm_class_reply(globalconf.connection,
@@ -213,9 +238,9 @@ property_update_wm_class(client_t *c, xcb_get_property_cookie_t cookie)
 				     &hint, NULL))
         return;
 
-    luaA_object_push(globalconf.L, c);
-    client_set_class_instance(globalconf.L, -1, hint.class_name, hint.instance_name);
-    lua_pop(globalconf.L, 1);
+    luaA_object_push(L, c);
+    client_set_class_instance(L, -1, hint.class_name, hint.instance_name);
+    lua_pop(L, 1);
 
     xcb_icccm_get_wm_class_reply_wipe(&hint);
 }
@@ -241,13 +266,14 @@ property_get_net_wm_icon(client_t *c)
 void
 property_update_net_wm_icon(client_t *c, xcb_get_property_cookie_t cookie)
 {
-    cairo_surface_t *surface = ewmh_window_icon_get_reply(cookie);
-
-    if(!surface)
+    cairo_surface_array_t array = ewmh_window_icon_get_reply(cookie);
+    if (array.len == 0)
+    {
+        cairo_surface_array_wipe(&array);
         return;
-
-    client_set_icon(c, surface);
-    cairo_surface_destroy(surface);
+    }
+    c->have_ewmh_icon = true;
+    client_set_icons(c, array);
 }
 
 xcb_get_property_cookie_t
@@ -268,9 +294,10 @@ property_update_net_wm_pid(client_t *c, xcb_get_property_cookie_t cookie)
         uint32_t *rdata = xcb_get_property_value(reply);
         if(rdata)
         {
-            luaA_object_push(globalconf.L, c);
-            client_set_pid(globalconf.L, -1, *rdata);
-            lua_pop(globalconf.L, 1);
+            lua_State *L = globalconf_get_lua_State();
+            luaA_object_push(L, c);
+            client_set_pid(L, -1, *rdata);
+            lua_pop(L, 1);
         }
     }
 
@@ -306,8 +333,6 @@ property_update_wm_protocols(client_t *c, xcb_get_property_cookie_t cookie)
 /** The property notify event handler.
  * \param state currently unused
  * \param window The window to obtain update the property with.
- * \param name The protocol atom, currently unused.
- * \param reply (Optional) An existing reply.
  */
 static int
 property_handle_xembed_info(uint8_t state,
@@ -322,7 +347,8 @@ property_handle_xembed_info(uint8_t state,
                              XCB_GET_PROPERTY_TYPE_ANY, 0, 3);
         xcb_get_property_reply_t *propr =
             xcb_get_property_reply(globalconf.connection, cookie, 0);
-        xembed_property_update(globalconf.connection, emwin, propr);
+        xembed_property_update(globalconf.connection, emwin,
+                               globalconf.timestamp, propr);
         p_delete(&propr);
     }
 
@@ -333,22 +359,23 @@ static int
 property_handle_net_wm_opacity(uint8_t state,
                                xcb_window_t window)
 {
+    lua_State *L = globalconf_get_lua_State();
     drawin_t *drawin = drawin_getbywin(window);
 
     if(drawin)
     {
-        luaA_object_push(globalconf.L, drawin);
-        window_set_opacity(globalconf.L, -1, xwindow_get_opacity(drawin->window));
-        lua_pop(globalconf.L, -1);
+        luaA_object_push(L, drawin);
+        window_set_opacity(L, -1, xwindow_get_opacity(drawin->window));
+        lua_pop(L, -1);
     }
     else
     {
         client_t *c = client_getbywin(window);
         if(c)
         {
-            luaA_object_push(globalconf.L, c);
-            window_set_opacity(globalconf.L, -1, xwindow_get_opacity(c->window));
-            lua_pop(globalconf.L, 1);
+            luaA_object_push(L, c);
+            window_set_opacity(L, -1, xwindow_get_opacity(c->window));
+            lua_pop(L, 1);
         }
     }
 
@@ -359,16 +386,56 @@ static int
 property_handle_xrootpmap_id(uint8_t state,
                              xcb_window_t window)
 {
-    signal_object_emit(globalconf.L, &global_signals, "wallpaper_changed", 0);
+    lua_State *L = globalconf_get_lua_State();
+    root_update_wallpaper();
+    signal_object_emit(L, &global_signals, "wallpaper_changed", 0);
     return 0;
 }
 
+/** The property notify event handler handling xproperties.
+ * \param ev The event.
+ */
+static void
+property_handle_propertynotify_xproperty(xcb_property_notify_event_t *ev)
+{
+    lua_State *L = globalconf_get_lua_State();
+    xproperty_t *prop;
+    xproperty_t lookup = { .atom = ev->atom };
+    buffer_t buf;
+    void *obj;
+
+    prop = xproperty_array_lookup(&globalconf.xproperties, &lookup);
+    if(!prop)
+        /* Property is not registered */
+        return;
+
+    if (ev->window != globalconf.screen->root)
+    {
+        obj = client_getbywin(ev->window);
+        if(!obj)
+            obj = drawin_getbywin(ev->window);
+        if(!obj)
+            return;
+    } else
+        obj = NULL;
+
+    /* Get us the name of the property */
+    buffer_inita(&buf, a_strlen(prop->name) + a_strlen("xproperty::") + 1);
+    buffer_addf(&buf, "xproperty::%s", prop->name);
+
+    /* And emit the right signal */
+    if (obj)
+    {
+        luaA_object_push(L, obj);
+        luaA_object_emit_signal(L, -1, buf.s, 0);
+        lua_pop(L, 1);
+    } else
+        signal_object_emit(L, &global_signals, buf.s, 0);
+    buffer_wipe(&buf);
+}
 
 /** The property notify event handler.
- * \param data Unused data.
- * \param connection The connection to the X server.
  * \param ev The event.
- * \return Status code, 0 if everything's fine.
  */
 void
 property_handle_propertynotify(xcb_property_notify_event_t *ev)
@@ -378,6 +445,8 @@ property_handle_propertynotify(xcb_property_notify_event_t *ev)
 
     globalconf.timestamp = ev->time;
 
+    property_handle_propertynotify_xproperty(ev);
+
     /* Find the correct event handler */
 #define HANDLE(atom_, cb) \
     if (ev->atom == atom_) \
@@ -421,4 +490,76 @@ property_handle_propertynotify(xcb_property_notify_event_t *ev)
     (*handler)(ev->state, ev->window);
 }
 
+/** Register a new xproperty.
+ * \param L The Lua VM state.
+ * \return The number of elements pushed on stack.
+ * \luastack
+ * \lparam The name of the X11 property
+ * \lparam One of "string", "number" or "boolean"
+ */
+int
+luaA_register_xproperty(lua_State *L)
+{
+    const char *name;
+    struct xproperty property;
+    struct xproperty *found;
+    const char *const args[] = { "string", "number", "boolean" };
+    xcb_intern_atom_reply_t *atom_r;
+    int type;
+
+    name = luaL_checkstring(L, 1);
+    type = luaL_checkoption(L, 2, NULL, args);
+    if (type == 0)
+        property.type = PROP_STRING;
+    else if (type == 1)
+        property.type = PROP_NUMBER;
+    else
+        property.type = PROP_BOOLEAN;
+
+    atom_r = xcb_intern_atom_reply(globalconf.connection,
+                                   xcb_intern_atom_unchecked(globalconf.connection, false,
+                                                             a_strlen(name), name),
+                                   NULL);
+    if(!atom_r)
+        return 0;
+
+    property.atom = atom_r->atom;
+    p_delete(&atom_r);
+
+    found = xproperty_array_lookup(&globalconf.xproperties, &property);
+    if(found)
+    {
+        /* Property already registered */
+        if(found->type != property.type)
+            return luaL_error(L, "xproperty '%s' already registered with different type", name);
+    }
+    else
+    {
+        property.name = a_strdup(name);
+        xproperty_array_insert(&globalconf.xproperties, property);
+    }
+
+    return 0;
+}
+
+/** Set an xproperty.
+ * \param L The Lua VM state.
+ * \return The number of elements pushed on stack.
+ */
+int
+luaA_set_xproperty(lua_State *L)
+{
+    return window_set_xproperty(L, globalconf.screen->root, 1, 2);
+}
+
+/** Get an xproperty.
+ * \param L The Lua VM state.
+ * \return The number of elements pushed on stack.
+ */
+int
+luaA_get_xproperty(lua_State *L)
+{
+    return window_get_xproperty(L, globalconf.screen->root, 1);
+}
+
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
