@@ -22,13 +22,15 @@
 #ifndef AWESOME_SPAWN_H
 #define AWESOME_SPAWN_H
 
-#include "globalconf.h"
+#include "objects/client.h"
+
+#include <lua.h>
 
 void spawn_init(void);
+void spawn_handle_reap(const char *);
+char * const * spawn_transform_commandline(char **);
 void spawn_start_notify(client_t *, const char *);
 int luaA_spawn(lua_State *);
 
-int spawn_system(const char *);
-
 #endif
 // vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
